@@ -13,12 +13,16 @@ pub fn make_into_heads(info: &StructInfo, options: Options) -> (TokenStream, Tok
     let mut code = Vec::new();
     let mut field_initializers = Vec::new();
     let mut head_fields = Vec::new();
+    let internal_struct = &info.internal_ident;
     // Drop everything in the reverse order of what it was declared in. Fields that come later
     // are only dependent on fields that came before them.
     for field in info.fields.iter().rev() {
         let field_name = &field.name;
-        if !field.self_referencing {
-            code.push(quote! { let #field_name = self.#field_name; });
+        if field.self_referencing {
+            // Heads are fields that do not borrow anything.
+            code.push(quote! { ::core::mem::drop(this.#field_name); });
+        } else {
+            code.push(quote! { let #field_name = this.#field_name; });
             if field.is_borrowed() {
                 field_initializers
                     .push(quote! { #field_name: ::ouroboros::macro_help::unbox(#field_name) });
@@ -27,9 +31,6 @@ pub fn make_into_heads(info: &StructInfo, options: Options) -> (TokenStream, Tok
             }
             let field_type = &field.typ;
             head_fields.push(quote! { #visibility #field_name: #field_type });
-        } else {
-            // Heads are fields that do not borrow anything.
-            code.push(quote! { ::core::mem::drop(self.#field_name); });
         }
     }
     for (ty, ident) in info.generic_consumers() {
@@ -71,6 +72,9 @@ pub fn make_into_heads(info: &StructInfo, options: Options) -> (TokenStream, Tok
         #[allow(clippy::drop_copy)]
         #[allow(clippy::drop_non_drop)]
         #visibility fn into_heads(self) -> Heads<#(#generic_args),*> {
+            let this_ptr = &self as *const _;
+            let this: #internal_struct<#(#generic_args),*> = unsafe { ::core::mem::transmute_copy(&*this_ptr) };
+            ::core::mem::forget(self);
             #(#code)*
             Heads {
                 #(#field_initializers),*
